__all__ = ['model_from_owl_str', 'model_from_owl_file', 'model_to_owl_str',
           'model_to_owl_file', 'model_from_owl_url', 'model_from_pc_query']

import requests
from xml.etree import ElementTree as ET
from .biopax.model import BioPaxModel
from .xml_util import xml_to_str, xml_to_file
from .pc_client import graph_query


def model_from_owl_str(owl_str):
    """Return a BioPAX Model from an OWL string.

    Parameters
    ----------
    owl_str : str
        A OWL string of BioPAX content.

    Returns
    -------
    pybiopax.biopax.BioPaxModel
        A BioPAX Model deserialized from the OWL string.
    """
    return BioPaxModel.from_xml(ET.fromstring(owl_str))


def model_from_owl_file(fname):
    """Return a BioPAX Model from an OWL string.

    Parameters
    ----------
    fname : str
        A OWL file of BioPAX content.

    Returns
    -------
    pybiopax.biopax.BioPaxModel
        A BioPAX Model deserialized from the OWL file.
    """
    with open(fname, 'r') as fh:
        owl_str = fh.read()
        return model_from_owl_str(owl_str)


def model_from_owl_url(url):
    """Return a BioPAX Model from an URL pointing to an OWL file.

    Parameters
    ----------
    url : str
        A OWL URL with BioPAX content.

    Returns
    -------
    pybiopax.biopax.BioPaxModel
        A BioPAX Model deserialized from the OWL file.
    """
    res = requests.get(url)
    res.raise_for_status()
    return model_from_owl_str(res.text)


def model_from_pc_query(kind, source, target=None, **query_params):
    """Return a BioPAX Model from a Pathway Commons query.

    For more information on these queries, see
    http://www.pathwaycommons.org/pc2/#graph

    Parameters
    ----------
    kind : str
        The kind of graph query to perform. Currently 3 options are
        implemented, 'neighborhood', 'pathsbetween' and 'pathsfromto'.
    source : list[str]
        A single gene name or a list of gene names which are the source set for
        the graph query.
    target : Optional[list[str]]
        A single gene name or a list of gene names which are the target set for
        the graph query. Only needed for 'pathsfromto' queries.
    limit : Optional[int]
        This limits the length of the longest path considered in
        the graph query. Default: 1
    organism : Optional[str]
        The organism used for the query. Default: '9606' corresponding
        to human.
    datasource : Optional[list[str]]
        A list of database sources that the query results should include.
        Example: ['pid', 'panther']. By default, all databases are considered.

    Returns
    -------
    pybiopax.biopax.BioPaxModel
        A BioPAX Model obtained from the results of the Pathway Commons query.
    """
    owl_str = graph_query(kind, source, target=target, **query_params)
    return model_from_owl_str(owl_str)


def model_to_owl_str(model):
    """Return an OWL string serialized from a BioPaxModel object.

    Parameters
    ----------
    model : pybiopax.biopax.BioPaxModel
        The BioPaxModel to serialize into an OWL string.

    Returns
    -------
    str
        The OWL string for the model.
    """
    return xml_to_str(model.to_xml())


def model_to_owl_file(model, fname):
    """Write an OWL string serialized from a BioPaxModel object into a file.

    Parameters
    ----------
    model : pybiopax.biopax.BioPaxModel
        The BioPaxModel to serialize into an OWL file.
    fname : str
        The path to the target OWL file.
    """
    xml_to_file(model.to_xml(), fname)



"""A client to the PathwayCommons REST API. For more details about
the service, see the documentation at https://www.pathwaycommons.org/pc2/."""
import logging
import requests


logger = logging.getLogger(__name__)


pc2_url = 'http://www.pathwaycommons.org/pc2/'


def graph_query(kind, source, target=None, **query_params):
    """Perform a graph query on PathwayCommons.

    For more information on these queries, see
    http://www.pathwaycommons.org/pc2/#graph

    Parameters
    ----------
    kind : str
        The kind of graph query to perform. Currently 3 options are
        implemented, 'neighborhood', 'pathsbetween' and 'pathsfromto'.
    source : list[str]
        A single gene name or a list of gene names which are the source set for
        the graph query.
    target : Optional[list[str]]
        A single gene name or a list of gene names which are the target set for
        the graph query. Only needed for 'pathsfromto' queries.
    limit : Optional[int]
        This limits the length of the longest path considered in
        the graph query. Default: 1
    organism : Optional[str]
        The organism used for the query. Default: '9606' corresponding
        to human.
    datasource : Optional[list[str]]
        A list of database sources that the query results should include.
        Example: ['pid', 'panther']. By default, all databases are considered.

    Returns
    -------
    str
        A BioPAX OWL string that can then be deserialized into a BioPaxModel.
    """

    params = {}
    params['format'] = 'BIOPAX'
    params['organism'] = query_params.get('organism', '9606')
    params['datasource'] = query_params.get('datasource', None)
    # Get the "kind" string
    kind_str = kind.lower()
    if kind not in ['neighborhood', 'pathsbetween', 'pathsfromto']:
        raise ValueError('Invalid query type %s' % kind_str)
    params['kind'] = kind_str
    # Get the source string
    params['source'] = _get_query_entity(source)
    try:
        params['limit'] = int(query_params.get('limit', 1))
    except (TypeError, ValueError):
        raise ValueError('Invalid neighborhood limit %s' %
                         query_params.get('limit'))
    params['target'] = _get_query_entity(target)

    logger.info('Sending Pathway Commons query with parameters: ')
    for k, v in params.items():
        logger.info(' %s: %s' % (k, v))

    res = requests.get(pc2_url + 'graph', params=params)
    if not res.status_code == 200:
        logger.error('Response is HTTP code %d.' % res.status_code)
        if res.status_code == 500:
            logger.error('Note: HTTP code 500 can mean empty '
                         'results for a valid query.')
        return None
    return res.text


def _get_query_entity(ent):
    if isinstance(ent, str):
        return ent
    elif isinstance(ent, (list, tuple)):
        return ','.join(ent)
    else:
        raise ValueError('Invalid query entity: %s' % str(ent))



"""This module implements the BioPAX Level 3 object model as
a set of classes with inheritance. At the top of the class
hierarchy is the generic BioPaxObject."""
from .base import *
from .interaction import *
from .physical_entity import *
from .util import *
from .model import *


import re
from lxml import etree
from lxml.builder import ElementMaker


namespaces = {
    'xsd': 'http://www.w3.org/2001/XMLSchema#',
    'owl': 'http://www.w3.org/2002/07/owl#',
    'rdf': 'http://www.w3.org/1999/02/22-rdf-syntax-ns#',
    'bp': 'http://www.biopax.org/release/biopax-level3.owl#',
    'xml': 'http://www.w3.org/XML/1998/namespace'
}


makers = {
    ns: ElementMaker(namespace=prefix)
    for ns, prefix in namespaces.items()
}


def wrap_xml_elements(elements):
    """Return a valid BioPAX OWL wrapping XML-serialized BioPAX objects."""
    # We first make the RDF wrapper and add an Ontology element first
    rdfm = ElementMaker(namespace=namespaces['rdf'],
                        nsmap=namespaces)
    rdf_element = rdfm('RDF',
                       **{nselem('xml', 'base'): 'http://purl.org/pc2/7/'})
    owl_element = makers['owl']('Ontology',
                                **{nselem('rdf', 'about'): ''})
    imports = makers['owl']('imports',
                            **{nselem('rdf', 'resource'): namespaces['bp']})
    owl_element.append(imports)
    rdf_element.append(owl_element)

    for element in elements:
        rdf_element.append(element)
    return rdf_element


def xml_to_str(xml):
    """Return the OWL string for an XML element tree."""
    xmlb = etree.tostring(xml, pretty_print=True,
                          encoding='utf-8', xml_declaration=True)
    xmls = xmlb.decode('utf-8')
    xmls = '\n'.join([re.sub(r'^  <bp', '\n<bp', x) for x in xmls.split('\n')])
    xmls = '\n'.join([re.sub(r'^  </bp', '</bp', x) for x in xmls.split('\n')])
    xmls = '\n'.join([re.sub(r'^    <', ' <', x) for x in xmls.split('\n')])
    return xmls


def xml_to_file(xml, fname):
    """Write an XML element tree to a given file."""
    with open(fname, 'w') as fh:
        fh.write(xml_to_str(xml))


def nselem(ns, elem):
    """Return a full namespaced string with curly brackets with a suffix."""
    return '{%s}%s' % (namespaces[ns], elem)


def nssuffix(ns, suffix):
    """Return a full namespaced string with a suffix."""
    return '%s%s' % (namespaces[ns], suffix)


def get_datatype(attrib):
    """Return the RDF data type of an element attribute."""
    return attrib.get(nselem('rdf', 'datatype'))


def get_resource(attrib):
    """Return the resource associated with an element attribute."""
    res = attrib.get(nselem('rdf', 'resource'))
    if res:
        if res.startswith('#'):
            return res[1:]
        else:
            return res


def is_url(txt):
    """Return true if the given string is an URL."""
    return txt.startswith('http')


def is_datatype(attrib, datatype):
    """Return True of the given attribute is of a given type."""
    return get_datatype(attrib) == datatype


def get_tag(element):
    """Return the tag of an element."""
    return re.match(r'.*}(.+)', element.tag).groups()[0]


def get_attr_tag(element):
    """Return the tag of an element as an attribute name."""
    raw_tag = get_tag(element)
    return camel_to_snake(raw_tag)


def get_id_or_about(element):
    """Return the ID or the about associated with an element"""
    return element.attrib.get(nselem('rdf', 'ID')) or \
        element.attrib.get(nselem('rdf', 'about'))


def get_ns(element):
    """Return the name space of a given element."""
    return re.match(r'\{(.*)\}', element.tag).groups()[0]


def has_ns(element, ns):
    """Return True if the element is from a given name space."""
    return get_ns(element) == namespaces[ns]


def camel_to_snake(txt):
    """Return snake case from camel case"""
    return re.sub(r'(?<!^)(?=[A-Z])', '_', txt).lower()


def snake_to_camel(txt):
    """Return camel case from snake case."""
    parts = txt.split('_')
    return parts[0] + ''.join([p.capitalize() for p in parts[1:]])



import re
from os import path
from setuptools import setup, find_packages


here = path.abspath(path.dirname(__file__))

with open(path.join(here, 'README.md'), encoding='utf-8') as fh:
    long_description = fh.read()

with open(path.join(here, 'pybiopax', '__init__.py'), 'r') as fh:
    for line in fh.readlines():
        match = re.match(r'__version__ = \'(.+)\'', line)
        if match:
            version = match.groups()[0]
            break
    else:
        raise ValueError('Could not get version from pybiopax/__init__.py')


setup(name='pybiopax',
      version=version,
      description=('A Python implementation of the BioPAX object model, '
                   'and parts of PaxTools.'),
      long_description=long_description,
      long_description_content_type='text/markdown',
      url='https://github.com/indralab/pybiopax',
      author='Benjamin M. Gyori, Harvard Medical School',
      author_email='benjamin_gyori@hms.harvard.edu',
      classifiers=[
          'Development Status :: 4 - Beta',
          'Programming Language :: Python :: 3.5',
          'Programming Language :: Python :: 3.6',
          'Programming Language :: Python :: 3.7'
      ],
      packages=find_packages(),
      install_requires=['lxml', 'requests'],
      extras_require={'test': ['nose', 'coverage']},
      keywords=['biology', 'pathway']
      )
